/***************************************************************************
 *   Copyright (C) 2007 by Chunkeey <chunkeey@web.de>                      *
 *   Copyright (C) 2008 by Jung-uk Kim <jkim@FreeBSD.org>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/types.h>

/* pidfile(3) */
#ifdef __FreeBSD__
#include <sys/param.h>
#include <libutil.h>
#endif

#include <usb.h>

#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_ttf.h>

#include "asusoled.h"
#include "asusoled_hw.h"

/* Version */
#define	MAJOR_VERSION		0
#define	MINOR_VERSION		10

/* pidfile(3) */
#ifdef __FreeBSD__
#ifndef DEFAULT_PID_FILE
#define	DEFAULT_PID_FILE	"/var/run/asusoled.pid"
#endif
#endif

/* Default TrueType font used for message rendering */
#ifndef DEFAULT_FONT
#define	DEFAULT_FONT		"/usr/local/lib/X11/fonts/TTF/luxisr.ttf"
#endif

/* Default font face for *.ttc */
#ifndef DEFAULT_FONT_INDEX
#define	DEFAULT_FONT_INDEX	0
#endif

/* Default font size in point (0 means auto-resize) */
#ifndef DEFAULT_FONT_SIZE
#define	DEFAULT_FONT_SIZE	0
#endif

/* Default daemon polling time in second */
#ifndef DEFAULT_POLLING_TIME
#define	DEFAULT_POLLING_TIME	60
#endif

/* Default USB timeout in millisecond */
#ifndef DEFAULT_USB_TIMEOUT
#define	DEFAULT_USB_TIMEOUT	1000
#endif

#ifndef BUFSIZ
#define	BUFSIZ			1024
#endif

static struct usb_dev_handle *handle;
static TTF_Font	*font;
#ifdef __FreeBSD__
static struct pidfh *pfh;
#endif

static void	cleanup_open_files(int);

static int	convert_image(unsigned char *, const unsigned char *,
		    const SDL_Rect *);
static void	resize_image(const SDL_Surface *, SDL_Rect *, int);
static unsigned char *recalc_image(const SDL_Surface *, int);

static TTF_Font	*open_font(const char *, int, int, int, long);

#ifdef __FreeBSD__
static int	process_kill(const char *);
#endif
static int	process_image(const char *, int);
static int	process_message(const char *, int, const char *,
		    int, int, long);
static int	process_command(const char *, int, const char *,
		    int, int, long, int);
static int	process_daemon(const char *, int, const char *,
		    int, int, long, int, struct usb_device *,
		    const char *, int);

static int	show_version(void);
static int	show_options(const char *);

struct usb_device *
oled_find_device(void)
{
	struct usb_device *dev;
	struct usb_bus *bus;

	usb_init();
	usb_find_busses();
	usb_find_devices();

	for (bus = usb_get_busses(); bus != NULL; bus = bus->next)
		for (dev = bus->devices; dev != NULL; dev = dev->next)
			if ((dev->descriptor.idVendor == 0x0b05) &&
			    (dev->descriptor.idProduct == 0x1726))
				return (dev);
	fprintf(stderr, "Could not find any ASUS LCM device!\n");

	return (NULL);
}

struct usb_dev_handle *
oled_open(struct usb_device *dev)
{
	struct usb_dev_handle *handle;

	if (dev == NULL || (handle = usb_open(dev)) == NULL) {
		fprintf(stderr, "Could not open USB device: %s\n",
		    strerror(errno));
		return (NULL);
	}

	if (usb_claim_interface(handle, 0) < 0) {
		/* LCM may be identified as a UHID device. */
#if defined(__FreeBSD__)
		fprintf(stderr, "\"kldunload uhid\" and try again.\n");
#elif defined(__linux__)
		fprintf(stderr, "\"rmmod usbhid\" and try again.\n");
#endif
		return (NULL);
	}

	return (handle);
}

int
oled_close(struct usb_dev_handle *handle)
{

	usb_release_interface(handle, 0);
	usb_close(handle);

	return (0);
}

int
oled_control(struct usb_dev_handle *handle, int mode)
{
	struct oled_packet buffer;

	switch (mode) {
	case DISPLAY_OFF:
		buffer.bitmap[0] = 0xae;
		break;
	case DISPLAY_ON:
		buffer.bitmap[0] = 0xaf;
		break;
	case DISPLAY_BLANK:
		return (oled_draw_image(handle, 0, 128, 32, NULL));
	case DISPLAY_FILL:
		return (oled_draw_image(handle, DRAW_NEGATE, 128, 32, NULL));
	default:
		return (EINVAL);
	}

	FILL_HEADER(buffer.header, 0x20, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00);
	if (usb_bulk_write(handle, 2, (char *)&buffer,
	    sizeof(buffer.header) + 1, DEFAULT_USB_TIMEOUT) < 0)
		return (errno);
	usleep(1000);

	return (0);
}

int
oled_draw_image(struct usb_dev_handle *handle, int func, int width, int height,
    const unsigned char *bitmap)
{
	SDL_Rect	rect;
	struct oled_packet *buf = calloc(width / 64, sizeof(*buf));
	int		i, fill;

	if (bitmap != NULL)
		for (i = 0; i < (width / 128) * (height / 16); i++) {
			switch (DRAW_EFFECT(func)) {
			default:
			case DRAW_STATIC:
				FILL_HEADER(buf[i].header, 0x10 + i, 0x80,
				    0x01, 0x01, 0x00, 0x01, 0x00);
				break;
			case DRAW_FLASH:
				FILL_HEADER(buf[i].header, 0x10, 0x80,
				    0x01, 0x01, 0x00, 0x00, 0xff);
				break;
			case DRAW_ROLL:
				FILL_HEADER(buf[i].header, 0x40, 0x80,
				    width / 64, i + 1, 0x00, 0x01, 0xff);
				break;
			}
			rect.x = (i / 2) * 128;
			rect.y = (i % 2) * 16;
			rect.w = width;
			convert_image(buf[i].bitmap, bitmap, &rect);
		}
	else {
		fill = (DRAW_CONVERT(func) & DRAW_NEGATE) == 0 ? 0xff : 0x00;
		memset(buf[0].bitmap, fill, sizeof(buf[0].bitmap));
		memset(buf[1].bitmap, fill, sizeof(buf[1].bitmap));
		FILL_HEADER(buf[0].header, 0x10, 0x80, 0x01, 0x01, 0x00,
		    0x01, 0x00);
		FILL_HEADER(buf[1].header, 0x11, 0x80, 0x01, 0x01, 0x00,
		    0x01, 0x00);
	}

	for (i = 0; i < width / 64; i++) {
		if (usb_bulk_write(handle, 2, (char *)&buf[i],
		    sizeof(buf[i]), DEFAULT_USB_TIMEOUT) < 0) {
			free(buf);
			return (errno);
		}
		usleep(1000);
	}
	free(buf);

	return (0);
}

static void
cleanup_open_files(int arg)
{

	/* Daemon is being refreshed, do nothing. */
	if (arg == SIGHUP)
		return;

	if (font != NULL)
		TTF_CloseFont(font);
#ifdef __FreeBSD__
	if (pfh != NULL)
		pidfile_remove(pfh);
#endif
	if (handle != NULL)
		oled_close(handle);

	if (arg != 0) {
		syslog(LOG_INFO, "Daemon terminated, pid: %jd",
		    (intmax_t)getpid());
		exit (0);
	}
}

static int
convert_image(unsigned char *dest, const unsigned char *src,
    const SDL_Rect *rect)
{
	int		i, x, y, z;
	unsigned char	buf;

	for (y = 0; y < 16; y += 8)
		for (x = 0; x < 128; x++) {
			for (z = 0, buf = 0; z < 8; z++) {
				i = (y + rect->y + z) * rect->w + x + rect->x;
				buf |= (src[i] != 0 ? 1 : 0) << z;
			}
			dest[127 - x + ((y >> 3) << 7)] = buf;
		}

	return (0);
}

static void
resize_image(const SDL_Surface *image, SDL_Rect *rect, int func)
{

	switch (DRAW_EFFECT(func)) {
	default:
	case DRAW_STATIC:
		rect->w = 128;
		rect->h = 32;
		break;
	case DRAW_FLASH:
		rect->w = 128;
		rect->h = 16;
		break;
	case DRAW_ROLL:
		if (image->w > 1792)
			rect->w = 1792;
		else {
			/* Round up to multiple of 128. */
			rect->w = (image->w % 128) == 0 ? image->w :
			    ((image->w / 128) + 1) * 128;
		}
		rect->h = 32;
		break;
	}
	/* Align center. */
	rect->x = (rect->w - image->w) / 2;
	rect->y = (rect->h - image->h) / 2;
}

static unsigned char *
recalc_image(const SDL_Surface *image, int func)
{
	SDL_Color	color;
	SDL_Rect	rect;
	unsigned char	*dst, *src, c;
	Uint32		pixel;
	int		i, bpp, x, y;

	resize_image(image, &rect, func);
	if ((dst = malloc(rect.w * rect.h)) == NULL) {
		fprintf(stderr, "Out of memory!\n");
		return (NULL);
	}

	src = (unsigned char *)image->pixels;
	bpp = image->format->BytesPerPixel;
	for (y = 0; y < rect.h; y++)
		for (x = 0; x < rect.w; x++) {
			if (x >= rect.x && x < (rect.x + image->w) &&
			    y >= rect.y && y < (rect.y + image->h)) {
				/* Get the pixel color in RGB. */
				pixel = 0;
				i = (y - rect.y) * image->pitch +
				    (x - rect.x) * bpp;
				memcpy(&pixel, src + i, bpp);
				SDL_GetRGB(pixel, image->format,
				    &color.r, &color.g, &color.b);
				/* Average dithering */
				if ((color.r + color.g + color.b) / 3 >= 0x80)
					c = 0x00;
				else
					c = 0xff;
			} else
				c = 0x00;

			/* Calculate index by the given effect. */
			if ((DRAW_CONVERT(func) & DRAW_FLIP) == 0)
				i = (DRAW_CONVERT(func) & DRAW_REVERSE) == 0 ?
				    (y * rect.w + x) : ((y + 1) * rect.w - x);
			else
				i = (DRAW_CONVERT(func) & DRAW_REVERSE) == 0 ?
				    ((rect.h - y - 1) * rect.w + x) :
				    ((rect.h - y) * rect.w - x);

			if ((DRAW_CONVERT(func) & DRAW_NEGATE) == 0)
				dst[i] = c;
			else
				dst[i] = ~c;
		}

	return (dst);
}

static TTF_Font *
open_font(const char *file, int func, int style, int size, long index)
{
	int		h;

	/* Do not load the font twice if already loaded. */
	if (font != NULL)
		return (font);

	/* Size 0 means auto-resize. */
	if (size == 0) {
		h = size = (DRAW_EFFECT(func) & DRAW_FLASH) == 0 ? 32 : 16;
		for (;;) {
			font = TTF_OpenFontIndex(file, size, index);
			if (font == NULL || h >= TTF_FontHeight(font))
				break;
			TTF_CloseFont(font);
			size--;
		}
	} else
		font = TTF_OpenFontIndex(file, size, index);

	if (font != NULL)
		TTF_SetFontStyle(font, style);

	return (font);
}

#ifdef __FreeBSD__
static int
process_kill(const char *pfile)
{
	pid_t		pid;

	pid = 0;
	pfh = pidfile_open(pfile, 0600, &pid);
	if (pfh == NULL) {
		if (errno == EEXIST && pid > 0 &&
		    kill(pid, SIGTERM) < 0)
			fprintf(stderr, "%s\n", strerror(errno));
		return (errno);
	} else {
		fprintf(stderr, "Daemon not found!\n");
		pidfile_remove(pfh);
		return (ESRCH);
	}
}
#endif

static int
process_image(const char *file, int func)
{
	SDL_Rect	rect;
	SDL_Surface	*image;
	unsigned char	*bitmp;
	int		ret;

	if ((image = IMG_Load(file)) == NULL) {
		fprintf(stderr, "%s\n", IMG_GetError());
		return (errno);
	}
	if ((bitmp = recalc_image(image, func)) == NULL) {
		ret = ENOMEM;
		fprintf(stderr, "Out of memory!\n");
		goto fail;
	}
	if ((ret = oled_control(handle, DISPLAY_BLANK)) != 0) {
		fprintf(stderr, "Failed to blank display!\n");
		goto fail;
	}
	resize_image(image, &rect, func);
	ret = oled_draw_image(handle, func, rect.w, rect.h, bitmp);
	free(bitmp);

fail:
	SDL_FreeSurface(image);

	return (ret);
}

static int
process_message(const char *msg, int func, const char *ffile, int style,
    int size, long index)
{
	SDL_Rect	rect;
	SDL_Color	black = {0x00, 0x00, 0x00};
	SDL_Color	white = {0xff, 0xff, 0xff};
	SDL_Surface	*image;
	unsigned char	*bitmp;
	int		ret;

	if (msg != NULL && msg[0] == '\0') {
		fprintf(stderr, "Nothing to display!\n");
		return (EINVAL);
	}
	if (TTF_Init() < 0) {
		fprintf(stderr, "%s\n", TTF_GetError());
		return (errno);
	}
	font = open_font(ffile, func, style, size, index);
	if (font == NULL) {
		fprintf(stderr, "%s\n", TTF_GetError());
		return (errno);
	}
	image = TTF_RenderUTF8_Shaded(font, msg, black, white);
	if ((bitmp = recalc_image(image, func)) == NULL) {
		ret = ENOMEM;
		fprintf(stderr, "Out of memory!\n");
		goto fail;
	}
	resize_image(image, &rect, func);
	ret = oled_draw_image(handle, func, rect.w, rect.h, bitmp);
	free(bitmp);

fail:
	SDL_FreeSurface(image);

	return (ret);
}

static int
process_command(const char *cmd, int func, const char *ffile, int style,
    int size, long index, int dflags)
{
	static	char	cmd_buf[BUFSIZ] = "";
	char		buf[BUFSIZ];
	FILE		*file;
	int		c, i;

	file = popen(cmd, "r");
	if (file == NULL) {
		fprintf(stderr, "Failed to execute command!\n");
		return (ENOENT);
	}

	i = 0;
	while (i < BUFSIZ - 1 && (c = fgetc(file)) != EOF) {
		if (isspace(c) != 0) {
			/* Skip leading white spaces. */
			if (i == 0)
				continue;
			/* We cannot handle multiple lines. */
			if (c == '\n')
				break;
			/* Replace other white spaces with ' '. */
			c = ' ';
		}
		buf[i++] = c;
	}
	pclose(file);

	if (i == 0) {
		fprintf(stderr, "Nothing returned by the command!\n");
		return (EINVAL);
	}
	buf[i] = '\0';

	/* If the message is the same, do not refresh in daemon mode. */
	if (dflags != 0) {
		if (strcmp(cmd_buf, buf) == 0)
			return (0);
		strcpy(cmd_buf, buf);
	}

	return (process_message(buf, func, ffile, style, size, index));
}

static int
process_daemon(const char *cmd, int func, const char *ffile, int style,
    int size, long index, int dflags, struct usb_device *device,
    const char *pfile, int poll)
{
#ifdef __FreeBSD__
	pid_t		pid;
#endif
	int		ret;

#ifdef __FreeBSD__
	/* Check if another daemon is running. */
	pfh = pidfile_open(pfile, 0600, &pid);
	if (pfh == NULL) {
		if (errno == EEXIST) {
			ret = errno;
			fprintf(stderr, "Daemon already running, pid: %jd\n",
			    (intmax_t)pid);
			goto done;
		}
		fprintf(stderr, "Cannot open or create pidfile\n");
	}
#endif

	/* Catch signals. */
	signal(SIGHUP, cleanup_open_files);
	signal(SIGTERM, cleanup_open_files);

	/* Daemonize and run in background. */
	if (daemon(0, 0) < 0) {
		ret = errno;
		fprintf(stderr, "Failed to daemonize!\n");
		goto done;
	}
	syslog(LOG_INFO, "Daemon started, pid: %jd", (intmax_t)getpid());

#ifdef __FreeBSD__
	/* Write pidfile if opened successfully. */
	if (pfh != NULL)
		pidfile_write(pfh);
#endif

	/* Daemon loops forever. */
	for (;;) {
		if ((ret = process_command(cmd, func, ffile, style, size,
		    index, dflags)) != 0)
			break;
		oled_close(handle);
		sleep(poll);
		handle = oled_open(device);
		if (handle == NULL) {
			ret = errno;
			break;
		}
	}
	syslog(LOG_ERR, "Daemon stopped: %s", strerror(ret));

done:
	cleanup_open_files(0);

	return (ret);
}

static int
show_version(void)
{

	printf("ASUS USB OLED Utility %d.%d\n", MAJOR_VERSION, MINOR_VERSION);
	return (0);
}

static int
show_options(const char *file)
{
	char		buf[16];

	printf("Usage:\t%s [-B|-b|-d|-e|-h|-V]\n", file);
	printf("\t%s -f image [-F|-R] [-n] [-r] [-v]\n", file);
	printf("\t%s -m message [opts] [-i] [-u] [-w] [-I index] "
	    "[-s size] [-t font]\n", file);
#ifdef __FreeBSD__
	printf("\t%s -c command [opts] [-D [-P time] [-p pidfile]]\n", file);
	printf("\t%s -k [-p pidfile]\n", file);
#else
	printf("\t%s -c command [opts] [-D [-P time]]\n", file);
#endif
	printf("\nControl Options:\n");
	printf("\t-B\t\tFill display.\n");
	printf("\t-b\t\tBlank display.\n");
	printf("\t-d\t\tDisable display.\n");
	printf("\t-e\t\tEnable display.\n");
	printf("\nOutput Options:\n");
	printf("\t-f image\tDisplay this image file.\n");
	printf("\t-m message\tDisplay this message.\n");
	printf("\t-c command\tDisplay output of this command.\n");
	printf("\nDisplay Options:\n");
	printf("\t-F\t\tDisplay image with flashing effect.\n");
	printf("\t-R\t\tDisplay image with rolling effect.\n");
	printf("\t-n\t\tDisplay negative image.\n");
	printf("\t-r\t\tDisplay horizontally reversed image.\n");
	printf("\t-v\t\tDisplay vertically flipped image.\n");
	printf("\nFont Options:\n");
	printf("\t-i\t\tSlant characters.\n");
	printf("\t-u\t\tUnderline characters.\n");
	printf("\t-w\t\tEmbolden characters.\n");
	printf("\t-I index\tLoad typeface of this index from TTC file.\n");
	printf("\t\t\t(default: %d)\n", DEFAULT_FONT_INDEX);
	printf("\t-s size\t\tSet nominal height of the font.\n");
	if (DEFAULT_FONT_SIZE != 0)
		snprintf(buf, sizeof(buf), "%d point(s)", DEFAULT_FONT_SIZE);
	printf("\t\t\t(default: %s)\n", DEFAULT_FONT_SIZE == 0 ? "auto" : buf);
	printf("\t-t font\t\tUse this font file.\n");
	printf("\nDaemon Options:\n");
	printf("\t-D\t\tRun the command continuously in background.\n");
	printf("\t-P time\t\tSet poll timer between command executions.\n");
	printf("\t\t\t(default: once per %u second(s))\n", DEFAULT_POLLING_TIME);
#ifdef __FreeBSD__
	printf("\t-p pidfile\tUse this pid file.\n");
	printf("\t\t\t(default: %s)\n", DEFAULT_PID_FILE);
	printf("\t-k\t\tStop running daemon.\n");
#endif
	printf("\nHelp Options:\n");
	printf("\t-h\t\tShow this usage message.\n");
	printf("\t-V\t\tShow version.\n");

	return (0);
}

#define	OPT_IMAGE	0x01
#define	OPT_FONT	0x02
#define	OPT_DAEMON	0x04
#define	OPT_PIDFILE	0x08

int
main(int argc, char *argv[])
{
	struct usb_device *device;
	char		*cmd, *msg, *ffile, *ifile, *pfile, *prog;
	long		index;
	int		cflags, dflags, eflags, fflags, kflags;
	int		fsize, mode, opt, opt_mask, poll, ret;

	errno = 0;
	cmd = msg = ffile = ifile = pfile = NULL;
	prog = basename(argv[0]);
	index = -1;
	cflags = dflags = eflags = fflags = kflags = mode = opt_mask = 0;
	poll = fsize = -1;

#ifdef __FreeBSD__
#define	OPT_LIST	"Bbc:DdeFf:hI:ikm:nP:p:Rrs:t:uVvw"
#else
#define	OPT_LIST	"Bbc:DdeFf:hI:im:nP:Rrs:t:uVvw"
#endif
	while ((opt = getopt(argc, argv, OPT_LIST)) != -1)
		switch (opt) {
		case 'B':
			cflags |= DISPLAY_FILL;
			mode++;
			break;
		case 'b':
			cflags |= DISPLAY_BLANK;
			mode++;
			break;
		case 'c':
			cmd = optarg;
			mode++;
			break;
		case 'D':
			dflags++;
			opt_mask |= OPT_DAEMON;
			break;
		case 'd':
			cflags |= DISPLAY_OFF;
			mode++;
			break;
		case 'e':
			cflags |= DISPLAY_ON;
			mode++;
			break;
		case 'F':
			eflags |= DRAW_FLASH;
			opt_mask |= OPT_IMAGE;
			break;
		case 'f':
			ifile = optarg;
			mode++;
			break;
		case 'h':
			return (show_options(prog));
			break;
		case 'I':
			index = (long)strtoul(optarg, NULL, 10);
			if ((index == 0 && errno != 0) || index < 0)
				return (show_options(prog));
			opt_mask |= OPT_FONT;
			break;
		case 'i':
			fflags |= TTF_STYLE_ITALIC;
			opt_mask |= OPT_FONT;
			break;
#ifdef __FreeBSD__
		case 'k':
			kflags++;
			mode++;
			break;
#endif
		case 'm':
			msg = optarg;
			mode++;
			break;
		case 'n':
			eflags |= DRAW_NEGATE;
			opt_mask |= OPT_IMAGE;
			break;
		case 'P':
			poll = (int)strtoul(optarg, NULL, 10);
			if ((poll == 0 && errno != 0) || poll < 0)
				return (show_options(prog));
			opt_mask |= OPT_DAEMON;
			break;
#ifdef __FreeBSD__
		case 'p':
			pfile = optarg;
			opt_mask |= OPT_PIDFILE;
			break;
#endif
		case 'R':
			eflags |= DRAW_ROLL;
			opt_mask |= OPT_IMAGE;
			break;
		case 'r':
			eflags |= DRAW_REVERSE;
			opt_mask |= OPT_IMAGE;
			break;
		case 's':
			fsize = (int)strtoul(optarg, NULL, 10);
			if ((fsize == 0 && errno != 0) || fsize < 0)
				return (show_options(prog));
			opt_mask |= OPT_FONT;
			break;
		case 't':
			ffile = optarg;
			opt_mask |= OPT_FONT;
			break;
		case 'u':
			fflags |= TTF_STYLE_UNDERLINE;
			opt_mask |= OPT_FONT;
			break;
		case 'V':
			return (show_version());
			break;
		case 'v':
			eflags |= DRAW_FLIP;
			opt_mask |= OPT_IMAGE;
			break;
		case 'w':
			fflags |= TTF_STYLE_BOLD;
			opt_mask |= OPT_FONT;
			break;
		default:
			show_options(prog);
			return (EINVAL);
		}

	argc -= optind;
	argv += optind;

	/* Check sanity of the options. */
	if (mode != 1 || (cflags != 0 && opt_mask != 0) ||
	    (kflags != 0 && (opt_mask & ~OPT_PIDFILE) != 0) ||
	    (ifile != NULL && (opt_mask & ~OPT_IMAGE) != 0) ||
	    (msg != NULL && (opt_mask & ~(OPT_IMAGE | OPT_FONT)) != 0) ||
	    (cmd != NULL && dflags == 0 &&
	    (opt_mask & (OPT_DAEMON | OPT_PIDFILE)) != 0) ||
	    DRAW_EFFECT(eflags) == (DRAW_FLASH | DRAW_ROLL)) {
		return (show_options(prog));
	}

	/* Look up LCM device and try opening it. */
	if ((device = oled_find_device()) == NULL)
		return (ENOENT);
	if ((handle = oled_open(device)) == NULL)
		return (ENOENT);

	/* Set default values. */
	if (ffile == NULL)
		ffile = DEFAULT_FONT;
	if (index == -1)
		index = DEFAULT_FONT_INDEX;
	if (fsize == -1)
		fsize = DEFAULT_FONT_SIZE;
#ifdef __FreeBSD__
	if (pfile == NULL)
		pfile = DEFAULT_PID_FILE;
#endif
	if (poll == -1)
		poll = DEFAULT_POLLING_TIME;

	/* Process command! */
	if (cflags != 0)
		ret = oled_control(handle, cflags);
#ifdef __FreeBSD__
	else if (kflags != 0)
		ret = process_kill(pfile);
#endif
	else if (ifile != NULL)
		ret = process_image(ifile, eflags);
	else if (msg != NULL)
		ret = process_message(msg, eflags, ffile, fflags, fsize,
		    index);
	else if (dflags == 0)
		ret = process_command(cmd, eflags, ffile, fflags, fsize,
		    index, dflags);
	else
		ret = process_daemon(cmd, eflags, ffile, fflags, fsize,
		    index, dflags, device, pfile, poll);

	cleanup_open_files(0);

	return (ret);
}
